/*
 * UnregisterServer() programmatically unregisters an in-proc COM
 * server, given its CLSID.
 *
 */
#include <windows.h>
#include <atlbase.h>  // for the string conversion funs. only
#include <tchar.h>
#include <stdio.h>
#include <malloc.h>

#include "UnregisterServer.h"

// number of chars needed to hold text rep. {xxxx-xx-xx-xx-xxxx}
// of a CLSID.
#define CLSID_STRLEN 38


typedef enum {
    InProcServer,
    InProcHandler,
    LocalServer
} ServerKind;

//
// FUNCTION:   CLSIDToBinaryPath(CLSID* rclsid, CTSTR* compKind)
//
// PURPOSE:    Given a CLSID, resolve it to the binary implementing
//             that component. 
//
// COMMENTS:   The responsibility of releasing the result string
//             is transferred to the caller. Release it using free().

LPTSTR
CLSIDToBinaryPath
    ( REFCLSID   rclsid
    , ServerKind sKind
    )
{
    USES_CONVERSION;
    HKEY        hKey;
    LONG        lRes;
    LPTSTR      subKeyStr;
    LPCTSTR     clsidPrefix = TEXT("CLSID");
    LPCTSTR     compKind;
    LPOLESTR    clsidString;
    int         subKeyStringLen;
    int         v;
    DWORD       dwType;
    DWORD       bufSize;
    LPTSTR      resString = NULL;
    HRESULT     hr;

    // construct the right registry key for the component

    switch (sKind) {
        // here's the reason why I miss having a CLSID -> binary component
        // resolver provided by the COM library - you have to emulate what
        // the SCM does (and, who knows, whether this will work with newer
        // versions of the SCM (e.g., for Win64)?)
    case InProcServer:  compKind = TEXT("InprocServer32"); break;
    case InProcHandler: compKind = TEXT("InprocHandler32"); break;
    case LocalServer:   compKind = TEXT("LocalServer32"); break;
    }

    subKeyStringLen = _tcslen(clsidPrefix) +
                      CLSID_STRLEN +
                      4 + _tcslen(compKind);

    subKeyStr = (LPTSTR)_alloca(sizeof(TCHAR) * subKeyStringLen);

    // convert the CLSID into "{xxxx-xx-xx-xx-xxxx}" form
    hr = StringFromCLSID(rclsid, &clsidString);

    if (FAILED(hr)) {
        return NULL;
    }

    v = _sntprintf(subKeyStr, subKeyStringLen, TEXT("%s\\%s\\%s"), clsidPrefix, OLE2T(clsidString), compKind);

    CoTaskMemFree(clsidString);

    if (v < 0) { // not enough room in subKeyStr
        //ToDo: use ReportErrorMessage() utility function.
        return NULL;
    }

    // use the registry to locate the DLL that implements
    // the server.

    lRes = RegOpenKeyEx(HKEY_CLASSES_ROOT, subKeyStr, 0, KEY_READ, &hKey);

    if (lRes != ERROR_SUCCESS) {
        return NULL;
    }


    // figure out the size of the key's unnamed value
    bufSize = 0; dwType = REG_SZ;
    lRes = RegQueryValueEx(hKey, NULL, NULL, &dwType, NULL, &bufSize);

    if (lRes == ERROR_SUCCESS) {
        // Q: does the 'bufSize' include the term. zero?
        resString = (LPTSTR)malloc(bufSize+1);

        lRes = RegQueryValueEx (hKey, NULL, NULL, &dwType, (LPBYTE)resString, &bufSize);

        if ( lRes != ERROR_SUCCESS || dwType != REG_SZ) {
            free(resString);
            resString = NULL;
        }
    }

    // close up and return
    RegCloseKey(hKey);
    return resString;

}

//
// FUNCTION: NukeRegEntry(REFCLSID)
//
// PURPOSE:  Given a CLSID, nuke the HKCR\CLSID\{clsid} registry sub-tree.
//           Poor man's unregistration, and is only used to ensure that
//           Registry is cleaned up the component's DllUnregisterServer()
//           method (well, NukeRegEntry() doesn't do all that a component
//           unregistration has to do.)
//           
HRESULT
NukeRegEntry
    (REFCLSID rclsid )
{
    USES_CONVERSION;
    LONG lRes;
    HRESULT hr;
    LPTSTR      subKeyStr;
    LPOLESTR    clsidString;
    int  subKeyStringLen;

    // convert the CLSID into "{xxxx-xx-xx-xx-xxxx}" form
    hr = StringFromCLSID(rclsid, &clsidString);

    subKeyStringLen = CLSID_STRLEN + 10;
    subKeyStr = (LPTSTR)_alloca(sizeof(TCHAR) * subKeyStringLen);

    _sntprintf(subKeyStr, subKeyStringLen, TEXT("CLSID\\%s"), OLE2T(clsidString));

    CoTaskMemFree(clsidString);

    lRes = SHDeleteKey(HKEY_CLASSES_ROOT, subKeyStr);

    return HRESULT_FROM_WIN32(lRes); 
}

//
// FUNCTION:    HRESULT UnregisterServer(CLSID* rclsid)
// 
// PURPOSE:     Given a CLSID, unregister the inproc server
//              associated with (if any) on the local machine.
//
// COMMENTS:    
HRESULT
UnregisterServer
    (REFCLSID rclsid )
{
    HRESULT hr = S_OK;

    // locate the path to the server.
    LPTSTR compPath = CLSIDToBinaryPath(rclsid, InProcServer);

    // mimic regsvr32 and call DllUnregisterServer() on the DLL.
    if (compPath != NULL) {
        HINSTANCE hDll = LoadLibrary(compPath);

        if (hDll != NULL) {
            HRESULT (STDAPICALLTYPE *pfn)(void);
            (FARPROC&)pfn = GetProcAddress(hDll, "DllUnregisterServer");

            if (pfn != NULL) {
                hr = pfn();
                if (FAILED(hr)) {
                    ;
                }
            }
            FreeLibrary(hDll);
            // too severe an action to perform without warning the user?
            // DeleteFile(compPath);
        }
    }
    hr = NukeRegEntry(rclsid);
    
    free(compPath);
    return hr;

}


//
// FUNCTION:    BinaryToCLSID()
//
// PURPOSE:     Given the path to a DLL, figure out its CLSID.
//
//              In general, this seems to be a hard problem, as there's
//              no way to get any info on what component(s) a DLL
//              implements other than asking it's class factory
//              to create an instance of a particular CLSID. Exhaustive
//              querying is not an
//
//              Looking beyond the binary itself, if we can assume the
//              component is already registered, we can do an exhaustive
//              search of the relevant portion of the Registry, and try
//              to find a match there. That's essentially what the function
//              does, but uses the Pan category to focus the search, only
//              checking installed Pan components. Mapping from CLSID to
//              a binary is done using the above CLSIDToBinaryPath().
//
//

// {E44A391E-8C72-4164-A82C-426E3541C40B}
static const GUID CATID_PanImage = 
{ 0xe44a391e, 0x8c72, 0x4164, { 0xa8, 0x2c, 0x42, 0x6e, 0x35, 0x41, 0xc4, 0xb } };

// matches the defn in PanObj/resource.h
#define ID_CLSID_STRING                 102


HRESULT
BinaryToCLSID( LPTSTR szFile,
               CLSID* pclsid,
               LPTSTR& effectName,
               LPTSTR& progID)
{
    USES_CONVERSION;

    HRESULT          hr;
    TCHAR            clsidString[100];
    HKEY             hKey;
    LONG             lr;
    DWORD            dwType = REG_SZ;
    DWORD            cbData = MAX_PATH - 1;


    HMODULE hMod = LoadLibrary(szFile);

    if ( LoadString(hMod, ID_CLSID_STRING, clsidString, 100) > 0) {
        CLSID clsid;
        TCHAR     buf[256];

        hr = CLSIDFromString(T2OLE(clsidString), &clsid);
        memcpy(pclsid,&clsid, sizeof(CLSID));

        // get hold of the ProgID and effect name from the
        // registry entry:

        _stprintf(buf, TEXT("CLSID\\%s"), clsidString);
        lr = RegOpenKeyEx(HKEY_CLASSES_ROOT,
                          buf,
                          0,
                          KEY_READ,
                          &hKey);

        cbData = MAX_PATH - 1;
        if ( lr == NO_ERROR ) {
            effectName = new TCHAR[MAX_PATH];

            lr = RegQueryValueEx(hKey, NULL, NULL, &dwType, (LPBYTE)effectName, &cbData);
            if ( lr == NO_ERROR ) {
                HKEY hKeyProgID;

                progID = new TCHAR[MAX_PATH];

                progID[0] = TEXT('\0');

                lr = RegOpenKeyEx (hKey, TEXT("VersionIndependentProgID"), 0, KEY_READ, &hKeyProgID);
                if (lr == NO_ERROR ) {
                   cbData = MAX_PATH - 1;

                   lr = RegQueryValueEx(hKeyProgID, NULL, NULL, &dwType, (LPBYTE)progID, &cbData);
                   RegCloseKey(hKeyProgID);
                }
            } else { 
                return E_FAIL;
            }
            return S_OK;
        }
    }
    return E_FAIL;
}

